'''
Defines the ReactiveComponent class to be inherited
to create reactive content
'''
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from ryzom.components import Component


class ReactiveComponent(Component):
    '''
    Subclass of component with reactive content.
    It takes a unique name and a ryzom.views.View in addition
    to common component parameters.

    :param str name: The (unique) name of the instance
    :param View view: The rizom View this instance is attached to
    '''
    def __init__(self, name, view, *content, **attrs):
        self.name = name
        self.view = view
        super().__init__(*content, **attrs)
        self.view.addReactiveComponent(self)

    def setcontent(self, content):
        '''
        This method should be called only by the associated view
        to update the component contents, then it sends the new
        content to the client associated with the view instance

        :param list(Component) content: A list of Component to replace \
                the current content of the instance
        '''
        self.content = content
        self.preparecontent()
        channel_name = self.view.channel_name
        if channel_name:
            channel = get_channel_layer()
            async_to_sync(channel.send)(channel_name, {
                'type': 'handle.ddp',
                'params': {
                    'type': 'changed',
                    'instance': self.to_obj()
                }
            })


class ReactiveDiv(ReactiveComponent):
    def __init__(self, name, view, *content):
        attrs = {
            'tag': 'div',
            '_id': f'reactive_div_{name}',
        }
        super().__init__(name, view, *content, **attrs)



from ryzom.components import Div, Ul, Li, Span, Input, Button


class Task(Li):
    def __init__(self, task_object):
        attrs = {'class': 'list-group-item'}
        content = [
            Span(task_object.about),
            Button('Done', **{'class': 'btn float-right'}, events={
                'click': f'ryzom.call("remove_task", {{id: {task_object.id}}})'
            })
        ]
        super().__init__(*content, **attrs, _id=f'task_{task_object.id}')


class TaskList(Ul):
    def __init__(self):
        self.subscriptions = ['task']
        super().__init__(**{'class': 'list-group'}, _id='tasklist')


class TaskForm(Div):
    def __init__(self):
        attrs = {
            'class': 'form-group',
            'style': 'margin:20px',
            'action': '#'
        }
        content = [
            Div(
                Input(**{
                    'id': 'task_input',
                    'type': 'text',
                    'class': 'form-control'
                }),
                Div(
                    Button(
                        'Add',
                        **{
                            'class': 'btn btn-primary'
                        },
                        events={
                            'click': ('ryzom.call("insert_task", '
                                      '{about: $("#task_input").value});'
                                      '$("#task_input").value = ""'
                                      )
                        }
                    ),
                    **{'class': 'input-group-append'}
                ),
                **{'class': 'input-group'}
            )
        ]
        super().__init__(*content, **attrs, _id='task_form')



import sys

from crudlfap.settings import *  # noqa


DATABASES['default']['ENGINE'] = 'django.db.backends.postgresql'
DATABASES['default']['NAME'] = 'ryzom'

# Find demo app modules when running under `ryzom runserver`
APP_DIR = os.path.dirname(os.path.abspath(__file__))
if APP_DIR not in sys.path:
    sys.path.insert(0, APP_DIR)

INSTALLED_APPS = [  # noqa: F405
    'ryz_ex',
    'todos',
    'channels',  # optional, but required for client-side interaction
    'ryzom',
    ] + INSTALLED_APPS + [  # noqa: F405
]

# CRUDLFA+ optional dependencies
OPTIONAL_APPS = [
    # {'debug_toolbar': {'after': 'django.contrib.staticfiles'}},
    {'django_extensions': {'before': 'crudlfap'}},
    {'collectdir': {'before': 'crudlfap'}},
]

OPTIONAL_MIDDLEWARE = [
    # {'debug_toolbar.middleware.DebugToolbarMiddleware': None}
]

MIDDLEWARE += [  # noqa: F405
    'ryzom.middleware.RyzomMiddleware',
]

if DEBUG:
    install_optional(OPTIONAL_APPS, INSTALLED_APPS)  # noqa: F405
    install_optional(OPTIONAL_MIDDLEWARE, MIDDLEWARE)  # noqa: F405

"""
AUTHENTICATION_BACKENDS += [  # noqa: F405
    'crudlfap_example.blog.crudlfap.AuthBackend',
]
"""

ROOT_URLCONF = 'ryz_ex.urls'

STATIC_ROOT = os.getenv(
    'STATIC_ROOT', Path(os.path.dirname(__file__)) / 'static')

ASGI_APPLICATION = 'ryz_ex.routing.application'
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [("localhost", 6379)],
        },
    },
}

# ryzom settings
RYZOM_APP = 'todos'
DDP_URLPATTERNS = 'todos.routing'
SERVER_METHODS = [
    'todos.methods'
]

CRUDLFAP_TEMPLATE_BACKEND["OPTIONS"]["globals"][  # noqa: F405
    "render_form"] = "ryz_ex.jinja2_ryzom.render_form"
#     "crudlfap.jinja2.render_form"

RYZOM_TEMPLATE_BACKEND = {
    "BACKEND": "ryzom.backends.ryzom.Ryzom",
    "OPTIONS": {
        "app_dirname": "components",
        "components_module": "ryzom.components.muicss",
        "components_prefix": "Mui",
        # "components_module": "ryzom.components.django",
        # "components_prefix": "Django",

        "context_processors": [
            "django.template.context_processors.debug",
            "django.template.context_processors.request",
            "django.contrib.auth.context_processors.auth",
            "django.contrib.messages.context_processors.messages",
            "django.template.context_processors.i18n",
        ],
        # "autoescape": True,
        # "auto_reload": DEBUG,
        # "translation_engine": "django.utils.translation",
        # "debug": False,
    }
}
TEMPLATES.append(RYZOM_TEMPLATE_BACKEND)

PYTEST_SKIP = os.getenv('PYTEST_SKIP', False)

"""
CRUDLFAP_TEMPLATE_BACKEND = {
    "BACKEND": "django_jinja.backend.Jinja2",
    "APP_DIRS": True,
    "OPTIONS": {
        "app_dirname": "jinja2",
        "match_extension": None,
        "context_processors": [
            "django.template.context_processors.debug",
            "django.template.context_processors.request",
            "django.contrib.auth.context_processors.auth",
            "django.contrib.messages.context_processors.messages",
            "django.template.context_processors.i18n",
        ],
        "extensions": [
            "jinja2.ext.do",
            "jinja2.ext.loopcontrols",
            "jinja2.ext.with_",
            "jinja2.ext.i18n",
            "jinja2.ext.autoescape",
            "django_jinja.builtins.extensions.CsrfExtension",
            "django_jinja.builtins.extensions.CacheExtension",
            "django_jinja.builtins.extensions.TimezoneExtension",
            "django_jinja.builtins.extensions.UrlsExtension",
            "django_jinja.builtins.extensions.StaticFilesExtension",
            "django_jinja.builtins.extensions.DjangoFiltersExtension",
        ],
        "constants": TEMPLATE_CONSTANTS,
        "globals": {
            "pagination_filter_params": "crudlfap.jinja2.pagination_filter_params",  # noqa
            "crudlfap_site": "crudlfap.site.site",
            "getattr": getattr,
            "str": str,
            "int": int,
            "isinstance": isinstance,
            "type": type,
            "render_form": "crudlfap.jinja2.render_form",
            "render_button": "bootstrap3.forms.render_button",
            "ryzom": "ryzom.components.component_html",
        },
        "newstyle_gettext": True,
        "bytecode_cache": {
            "name": "default",
            "backend": "django_jinja.cache.BytecodeCache",
            "enabled": False,
        },
        "autoescape": True,
        "auto_reload": DEBUG,
        "translation_engine": "django.utils.translation",
        "debug": False,
    }
}
"""



'''
Defines the django signals handlers.
'''
import importlib

from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from ryzom.ddp import send_insert, send_change, send_remove
from ryzom.models import Publications, Subscriptions
from ryzom.pubsub import Publishable


@receiver(post_save)
def _ddp_insert_change(sender, **kwargs):
    '''
    Function to send a DDP insert/change/remove messages to the channel layer
    whenever a Publishable model's save() method is called.
    This function will update the queryset of all subscriptions
    associated with the sender model and send insert, remove or change
    message for each _id that was added or removed from the old
    queryset to the new one.
    '''
    if Publishable not in sender.mro():
        return
    created = kwargs.pop('created')
    instance = kwargs.pop('instance')
    pubs = Publications.objects.filter(
            model_class=sender.__name__,
            model_module=sender.__module__)
    for pub in pubs:
        tmpl_module = importlib.import_module(pub.template_module)
        tmpl_class = getattr(tmpl_module, pub.template_class)
        subscriptions = Subscriptions.objects.filter(publication=pub)
        for sub in subscriptions:
            old_qs = sub.queryset
            new_qs = sub.exec_query(sender).aggregate(ids=ArrayAgg('id'))
            new_qs = new_qs['ids']
            # update the queryset
            sub.queryset = new_qs
            sub.save()

            diff = {
                'inserted': set(new_qs).difference(set(old_qs)),
                'removed': set(old_qs).difference(set(new_qs))
            }
            # if sets are the same
            if not diff['inserted'] and not diff['removed']:
                # if created and sets are the same,
                # entry has been filtered and can't be there
                if not created:
                    # changed and may have moved
                    # just send new instance and pos
                    send_change(sub, sender, tmpl_class, instance.id)

            # if sets aren't the same, then considering that only one entry
            # was added or has changed:
            # - it could have been removed if newly filtered (changed)
            # - it could have been added if no more filtered (changed)
            # - it could have been added if created and not filtered
            # - it cannot have just moved neither changed or the set
            #   would have been the same
            # - it could have been added while not created, replacing
            #   another entry because of filters, so created or not,
            #   we have to handle both added and removed entries
            else:
                # using loops for now but shouldn't be usefull as we
                # are handling only one entry, the queryset shouldn't
                # move by more that one in and/or one out
                for _id in diff['removed']:
                    send_remove(sub, sender, tmpl_class, _id)
                for _id in diff['inserted']:
                    send_insert(sub, sender, tmpl_class, _id)


@receiver(post_delete)
def _ddp_delete(sender, **kwargs):
    '''
    Function to send a DDP insert/remove messages to the channel layer
    whenever a Publishable model's delete() method is called.
    This function will update the queryset of all subscriptions
    associated with the sender model and send insert and remove
    message for each _id that was added or removed from the old
    queryset to the new one.
    '''
    if Publishable not in sender.mro():
        return
    instance = kwargs.pop('instance')
    pubs = Publications.objects.filter(
            model_module=sender.__module__,
            model_class=sender.__name__)
    for pub in pubs:
        tmpl_module = importlib.import_module(pub.template_module)
        tmpl_class = getattr(tmpl_module, pub.template_class)
        subscriptions = Subscriptions.objects.filter(publication=pub)
        for sub in subscriptions:
            old_qs = sub.queryset
            # if instance not in queryset, no need to remove it
            # or update the queryset
            if instance.id in old_qs:
                new_qs = sub.exec_query(sender).aggregate(ids=ArrayAgg('id'))
                new_qs = new_qs['ids']
                sub.queryset = new_qs
                sub.save()

                diff = {
                    'inserted': set(new_qs).difference(set(old_qs)),
                    'removed': set(old_qs).difference(set(new_qs))
                }

                for _id in diff['removed']:
                    send_remove(sub, sender, tmpl_class, _id)
                for _id in diff['inserted']:
                    send_insert(sub, sender, tmpl_class, _id)



from django.conf import settings
from django.db import models

from ryzom.pubsub import Publishable


class Task(Publishable, models.Model):
    user = models.ForeignKey(
        getattr(settings, 'AUTH_USER_MODEL', 'auth.User'),
        on_delete=models.SET_NULL, blank=True, null=True)
    about = models.CharField(max_length=1024)

    class Project:  # app_label - for app-specific model options
        ryzom = True

    def __str__(self):
        return self.about


Task.publish(
        name='task',
        template='todos.components.task.Task',
        query=[
            {'order_by': 'about'},
            {'offset': ('$count', -5)},
            {'limit': 5},
        ]
)



# Generated by Django 2.1.7 on 2019-03-07 15:31

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import ryzom.pubsub


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('about', models.CharField(max_length=1024)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(models.Model, ryzom.pubsub.Publishable),
        ),
    ]



from django.apps import AppConfig
from django.db import OperationalError, ProgrammingError


class RyzomConfig(AppConfig):
    '''
    Ryzom application configuration
    '''
    name = 'ryzom'

    def ready(self):
        '''
        AppConfig.ready overloading.
        Here we import the ryzom.signals module to connect
        the signals handlers to receivers,
        then we clean the Clients table (zombies clients stay after
        a server reboot)
        Finally, we create the publications that were registered
        in the to_publish list of ryzom.pusub module.
        '''
        try:
            import ryzom.signals  # noqa
        except (ImportError,):
            pass

        from ryzom.pubsub import to_publish

        try:
            Clients = self.get_model('Clients')
            Clients.objects.all().delete()
            for publication in to_publish:
                publication.publish_all()
        except (OperationalError, ProgrammingError):
            pass



from ryzom.components import Html, Head, Body, Meta, Title, Link, Static
from ryzom.components import Div, A, Nav, H1
from ryzom.reactive import ReactiveDiv

from .home import Welcome
from .task import TaskList, TaskForm


class Document(Html):
    def __init__(self, view):
        bs_src = 'ryz_ex/css/bootstrap.min.css'
        content = [
            Head(
                Title('TODOs'),
                Meta(**{'charset': 'utf-8'}),
                Link(**{'rel': 'stylesheet', 'href': Static(bs_src)})
            ),
            Body(Base(view))
        ]
        super().__init__(*content)


class Base(Div):
    def __init__(self, view):
        attrs = {
            'class': 'container',
            '_id': 'tasklist_container',
        }
        content = [
            A(Nav(H1('Todos'),
                  **{'class': 'navbar navbar-light bg-light'}),
              **{'href': '/'}
              ),
            ReactiveDiv('main-container', view, TaskList(), TaskForm()
                        ) if view.url == 'todos'
            else ReactiveDiv('main-container', view, Welcome())
        ]
        super().__init__(*content, **attrs)



'''
Ryzom components declarations.
There's still a lot of tags missing.
They will be added when they'll be needed
'''
import uuid

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.module_loading import import_string


def component_html(path, *args, **kwargs):
    from django.utils.safestring import mark_safe
    try:
        from jinja2.utils import Markup
    except ImportError:
        Markup = None
    ComponentCls = import_string(path)
    component = ComponentCls(*args, **kwargs)
    html = component.to_html()

    # DEBUG:
    if settings.DEBUG:
        print()
        print('HTML')
        print(html)
        print()

    if Markup:
        html = Markup(html)
    return mark_safe(html)


class Static:
    '''Return a static url for an app asset.

    :param str src: The app path of the asset.
    '''
    def __init__(self, src):
        self._data = src
        self.url = staticfiles_storage.url(src)

    def __str__(self):
        return self.url


class Component:
    '''Main ryzom component 'abstract' class to be inherited.

    This class defines the common attrsibutes and methods to all
    components,the main one being to_obj() that format an instance
    as a serializable dict that can be sent to the client over websocket
    the to_html() method is missing for now, it will be useful when
    implementing Server-Side Rendering

    Upon creation, if not specified, an instance of the Component class
    is considered as a child of the <html> tag, this is not guaranteed
    to be kept in a near future, because it's totally useless.
    Being a childnode of <body> seem a lot more meaningful.
    If no _id is specified, a random (but still unique) one will be
    generated.

    :param str tag: The HTML tag of the component
    :param list<Component> content: The component instances contained \
            by the current instance
    :param dict attrs: HTML attrsibutes (id, class, style, ...)
    :param dict events: The events to add listeners to \
            (click, hover, ...)
    :param str parent: The id of the component that contains the \
            current instance
    :param str _id: The _id of the current instance (must be unique)
    '''

    __subscriptions = []

    def __init__(self, *content, **attrs):
        if 'cls' in attrs:
            attrs['class'] = attrs.pop('cls')  # support HTML class attr

        self.content = list(content) or []

        self._id = attrs.pop('_id', uuid.uuid1().hex)
        self.parent = attrs.pop('parent', None)
        self.tag = attrs.pop('tag', 'div')
        self.events = attrs.pop('events', {})
        self.attrs = attrs or {}
        self.position = 0

        self.preparecontent()

    def preparecontent(self):
        '''Set the parent and position of children

        meant to be called by __init__().
        This method sets the current component as parent of each child
        Moreover it sets the child's position attribute to its index
        in the current component's content list
        '''
        # handle text node as content
        if self.tag == 'text':
            self.content = self.content[0]
        else:
            for i, c in enumerate(self.content):
                if not hasattr(c, 'to_html'):
                    c = Text(c)
                c.parent = self._id
                c.position = i

    def addchild(self, component):
        '''Add a child component

        Currently not used at all, but meant to push a new child
        at the end of the content's list
        A call to preparecontent() should follow the call to this method

        :param Component component: The child component to add to the \
                content of the current instance
        '''
        component.position = len(self.content)
        component.parent = self._id
        self.content.append(component)

    def addchildren(self, components):
        '''Add a list of children

        Currently not used at all, meant to push children at the
        end of the content's list.
        A call to preparecontent() should follow the call to this method

        :param list(Component) components: The component list to insert \
                in the content of the current instance
        '''
        for component in components:
            self.addchild(component)

    def addevents(self, events):
        '''Add events to instance

        Currently not used, meant to add/update a dict of eventListener
        attached to the DOM element associated with this instance

        :param dict events: The dict to update events with
        '''
        self.events.update(events)

    def to_obj(self):
        '''Get a serializable dict of the instance

        This methods returns a dict representation of the current
        instance. I handles subscriptions that will have this component
        instance _id as parent attribute.
        Recursively sets the content as dict too (maybe recursiveness is not
        a good thing to do without any control of how deep can the tree be,
        there's a risk of stack overflow that we must keep in mind)

        :returns: A serializable representation of the instance
        '''
        return {
            '_id': self._id,
            'tag': self.tag,
            'content': [
                c.to_obj()
                for c in self.content
            ] if self.tag != 'text' else self.content,
            'parent': self.parent,
            'position': self.position,
            'events': self.events,
            'attrs': self.attrs,
            'subscriptions': self.subscriptions
        }

    @property
    def subscriptions(self):
        return self.__subscriptions

    @subscriptions.setter
    def subscriptions(self, value=None):
        self.__subscriptions = value or []

    def to_html(self, context=None):
        if self.tag == 'text':
            return f'{self.content}'
        attrs = ''
        for k, v in self.attrs.items():
            if v is True:
                attrs += f'{k} '
            elif v is not False:
                attrs += f'{k}="{v}" '
        attrs += f'ryzom-id="{self._id}"'
        html = ''
        if getattr(self, 'selfclose', False):
            html = f'<{self.tag} {attrs}/>'
        elif getattr(self, 'noclose', False):
            html = f'<{self.tag} {attrs}>'
        else:
            html = f'<{self.tag} {attrs}>'
            for c in self.content:
                html += c.to_html() if getattr(c, 'to_html', None) else str(c)
            html += f'</{self.tag}>'
        return html

    def render(self, context=None):
        html = self.to_html(context)

        # DEBUG:
        if settings.DEBUG:
            print('template render HTML')
            print(html)

        return html


class Html(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'html'
        attrs['parent'] = None
        attrs['_id'] = 'html'
        super().__init__(*content, **attrs)


class Head(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'head'
        attrs['parent'] = 'html'
        attrs['_id'] = 'head'
        super().__init__(*content, **attrs)


class Body(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'body'
        attrs['parent'] = 'html'
        attrs['_id'] = 'body'
        super().__init__(*content, **attrs)


class Title(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'title'
        attrs['parent'] = 'head'
        attrs['_id'] = 'title'
        super().__init__(*content, **attrs)


class Meta(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'meta'
        attrs['parent'] = 'head'
        super().__init__(*content, **attrs)


class Link(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'link'
        attrs['parent'] = 'head'
        self.noclose = True
        super().__init__(*content, **attrs)


class Script(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'script'
        attrs['parent'] = 'head'
        super().__init__(*content, **attrs)


class Div(Component):
    '''
    Div component

    Represents a <div> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'div'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class A(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'a'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Ul(Component):
    '''
    Ul component

    Represents a <ul> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'ul'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Ol(Component):
    '''
    Ol component

    Represents a <ol> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'ol'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Li(Component):
    '''
    Li component

    Represents a <li> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'li'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Span(Component):
    '''
    Span component

    Represents a <span> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'span'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Text(Component):
    '''
    Text component

    Represents a text node

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'text'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Textarea(Component):
    '''
    Textarea component

    Represents a <textarea> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'textarea'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Form(Component):
    '''
    Form component

    Represents a <form> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'form'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Input(Component):
    '''
    Input component

    Represents a <input> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'input'
        attrs.setdefault('parent', 'body')
        self.noclose = True
        super().__init__(*content, **attrs)


class Select(Component):
    '''
    Select component

    Represents a <select> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'select'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Option(Component):
    '''
    Option component (within Select)

    Represents a <option> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'option'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Optgroup(Component):
    '''
    Option group component (within Select)

    Represents a <optgroup> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'optgroup'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Label(Component):
    '''
    Input component

    Represents a <input> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'label'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Button(Component):
    '''
    Button component

    Represents a <button> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'button'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class Icon(Component):
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'i'
        attrs.setdefault('parent', 'body')
        attrs['class'] = 'material-icons'
        super().__init__(*content, **attrs)


class Nav(Component):
    '''
    Nav component

    Represents a <nav> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'nav'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class H1(Component):
    '''
    H1 component

    Represents a <h1> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'h1'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class H2(Component):
    '''
    H2 component

    Represents a <h2> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'h2'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)


class H3(Component):
    '''
    H3 component

    Represents a <h3> HTML tag

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        content = content or []
        attrs = attrs or {}
        attrs['tag'] = 'h3'
        attrs.setdefault('parent', 'body')
        super().__init__(*content, **attrs)



# flake8: noqa: F401

from .components import (
    component_html,
    Static,
    Component,
    Html,
    Head,
    Body,
    Title,
    Meta,
    Link,
    Script,
    Div,
    A,
    Ul,
    Ol,
    Li,
    Span,
    Text,
    Textarea,
    Form,
    Input,
    Select,
    Option,
    Optgroup,
    Label,
    Button,
    Icon,
    Nav,
    H1,
    H2,
    H3
)



import importlib
from django.conf import settings


class RyzomMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        request.ryzom = None
        urls = importlib.import_module(settings.DDP_URLPATTERNS).urlpatterns
        url = request.path_info.lstrip('/')
        for u in urls:
            if u.pattern.match(url):
                request.ryzom = u.callback('')
                request.ryzom.oncreate(url)
                request.ryzom.onurl(url)
                break

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



'''
Routes channels websocket incoming message to ryzom.consumers.Consumer
'''
from django.conf.urls import url

from . import consumers

websocket_urlpatterns = [
    url(r'^ws/ddp/$', consumers.Consumer),
]
'''
The route to the consumer
'''



from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import sys


setup(
    name='ryzom',
    setup_requires='setupmeta',
    version='0.0.1',
    description='Meteorish Django responsive frontend',
    author='Thomas Mignot',
    author_email='jamespic@gmail.com',
    url='https://yourlabs.io/oss/ryzom',
    # packages=find_packages(),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    keywords='django frontend',
    install_requires=[
        'cli2',
        'django',
        'crudlfap',
    ],
    tests_require=[
        'tox',
    ],
    extras_require=dict(
        dev=[
        'channels',
        'channels-redis',
        #'psycopg2-binary',
        ],
    ),
    entry_points={
        'console_scripts': [
            'ryzom = ryzom_example.manage:main',
        ],
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3',
)
